import os
import sys

import pytest

sys.path.insert(1, os.path.join(sys.path[0], ".."))

import datetime

from orso.types import OrsoTypes

from orso.schema import FlatColumn
from orso.schema import FunctionColumn
from orso.schema import ConstantColumn
from orso.schema import DictionaryColumn

from orso.exceptions import ColumnDefinitionError


def test_flat_column_materialize():
    flat_column = FlatColumn(name="athena", type=OrsoTypes.INTEGER)

    with pytest.raises(TypeError):
        flat_column.materialize()

    assert flat_column.name == "athena"
    assert flat_column.type == OrsoTypes.INTEGER


def test_columns_with_unknown_parameters():
    FlatColumn(name="athena", type=OrsoTypes.INTEGER, alpha="betty")
    FunctionColumn(name="aries", type=OrsoTypes.DATE, binding=datetime.date.today, sketty="yum")


def test_column_type_mapping():
    fc = FlatColumn(name="athena", type="INTEGER")
    assert fc.type == OrsoTypes.INTEGER
    assert fc.type.__class__ == OrsoTypes

    fc = FlatColumn(name="athens", type="LIST")
    assert fc.type == OrsoTypes.ARRAY, fc.type

    fc = FlatColumn(name="athens", type="NUMERIC")
    assert fc.type == OrsoTypes.DOUBLE, fc.type

    fc = FlatColumn(name="athled", type=0)
    assert fc.type == 0

    with pytest.raises(ValueError):
        FlatColumn(name="able", type="LEFT")


def test_missing_columns():
    with pytest.raises(ColumnDefinitionError):
        FlatColumn(name="brian")


def test_type_checks():
    from decimal import Decimal
    from orso.schema import RelationSchema

    TEST_DATA = {
        OrsoTypes.VARCHAR: "string",
        OrsoTypes.INTEGER: 100,
        OrsoTypes.BOOLEAN: True,
        OrsoTypes.DATE: datetime.date.today(),
        OrsoTypes.ARRAY: ["a", "b", "c"],
        OrsoTypes.DOUBLE: 10.00,
        OrsoTypes.TIMESTAMP: datetime.datetime.utcnow(),
        OrsoTypes.TIME: datetime.time.min,
        OrsoTypes.BLOB: b"blob",
        OrsoTypes.DECIMAL: Decimal("3.7"),
        OrsoTypes.STRUCT: {"a": 1},
        OrsoTypes.INTERVAL: datetime.timedelta(days=1),
    }

    columns = []
    for t, v in TEST_DATA.items():
        columns.append(FlatColumn(name=str(t), type=t))

    schema = RelationSchema(name="test", columns=columns)
    schema.validate({str(k): v for k, v in TEST_DATA.items()})


def test_function_column():
    func_column = FunctionColumn(name="aries", type=OrsoTypes.DATE, binding=datetime.date.today)

    assert func_column.name == "aries"
    assert func_column.type == OrsoTypes.DATE
    assert func_column.length == 1
    assert func_column.aliases == []
    assert func_column.description == None

    values = func_column.materialize()

    assert len(values) == 1
    assert values[0] == datetime.date.today()

    func_column.length = 10
    values = func_column.materialize()

    assert len(values) == 10
    assert all(v == datetime.date.today() for v in values)


def test_constant_column():
    FACT: str = "The god of merchants, shepherds and messengers."

    const_column = ConstantColumn(name="hermes", type=OrsoTypes.VARCHAR, value=FACT)

    assert const_column.name == "hermes"
    assert const_column.type == OrsoTypes.VARCHAR
    assert const_column.value == FACT
    assert const_column.length == 1
    assert const_column.aliases == []
    assert const_column.description == None

    values = const_column.materialize()

    assert len(values) == 1
    assert values[0] == FACT

    const_column.length = 10
    values = const_column.materialize()

    assert len(values) == 10
    assert all(v == FACT for v in values)


def test_dict_column():
    MONTH_LENGTHS: list = ["31", "28", "31", "30", "31", "30", "31", "31", "30", "31", "30", "31"]

    dict_column = DictionaryColumn(name="pan", type=OrsoTypes.VARCHAR, values=MONTH_LENGTHS)

    assert dict_column.name == "pan"
    assert dict_column.type == OrsoTypes.VARCHAR

    assert sorted(dict_column.dictionary) == ["28", "30", "31"]
    assert list(dict_column.encoding) == [2, 0, 2, 1, 2, 1, 2, 2, 1, 2, 1, 2]

    values = dict_column.materialize()

    assert list(values) == MONTH_LENGTHS


if __name__ == "__main__":  # prgama: nocover
    from tests import run_tests

    run_tests()



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# ---------------------------------------------------------------------------------
# Our record when converted to bytes looks like this
#
#       ┌───────────┬─────────┬─────────────┬──────────────┐
#       │   flags   │  parity │   row_size  │  row values  │
#       └───────────┴─────────┴─────────────┴──────────────┘
#
# 'flags' currently has a a version string and a single flag of deleted.
# 'parity' is one byte, it is all of the bytes in the row values XORed
#
# Row object looks and acts like a Tuple where possible, but has additional features
# such as as_dict() to render as a dictionary.

from orso.exceptions import DataError

HEADER_SIZE: int = 6


def extract_columns(table, columns):
    # Initialize empty lists for each column
    result: tuple = tuple([] for _ in columns)

    # Extract the specified columns into the result lists
    for row in table:
        for j, column in enumerate(columns):
            result[j].append(row[column])

    return result


class Row(tuple):
    __slots__ = ()
    _fields: tuple = None

    def __new__(cls, data):
        return super().__new__(cls, data)

    @property
    def as_dict(self):
        return {k: v for k, v in zip(self._fields, self)}

    @property
    def values(self):
        return tuple(self)

    def keys(self):
        return self._fields

    def __repr__(self):
        return f"Row{super().__repr__()}"

    def __str__(self):
        return str(self.as_dict)

    def __setattr__(self, name, value):
        raise AttributeError("can't set attribute")

    def __delattr__(self, name):
        raise AttributeError("can't delete attribute")

    @classmethod
    def from_bytes(cls, data: bytes) -> tuple:
        import operator
        from functools import reduce

        from ormsgpack import unpackb

        if len(data) < HEADER_SIZE:
            raise DataError("Data malformed - missing bytes")

        if data[0] & 240 != 16:
            raise DataError("Data malformed - version error")

        record_bytes = data[HEADER_SIZE:]
        parity = reduce(operator.xor, record_bytes, 0)
        if parity != data[1]:
            raise DataError("Data malformed - parity check")
        record_size = int.from_bytes(data[2:HEADER_SIZE], byteorder="big")
        if len(record_bytes) != record_size:
            raise DataError("Data malformed - incorrect length")

        unpacked_values = unpackb(record_bytes)
        return cls(unpacked_values)

    def to_bytes(self) -> bytes:
        import operator
        from functools import reduce

        from ormsgpack import packb

        record_bytes = packb(tuple(self))
        parity = reduce(operator.xor, record_bytes, 0)
        record_size = len(record_bytes)
        if record_size > 16 * 1024 * 1024:
            raise DataError("Record length cannot exceed 16Mb")
        return b"\x10" + parity.to_bytes(1, "big") + record_size.to_bytes(4, "big") + record_bytes

    def to_json(self) -> bytes:
        import orjson

        return orjson.dumps(self.as_dict, default=str)

    @classmethod
    def create_class(cls, schema):
        from orso.schema import RelationSchema

        if isinstance(schema, RelationSchema):
            fields = tuple(c.name for c in schema.columns)
        else:
            fields = tuple(schema)
        return type("RowFactory", (Row,), {"_fields": fields})



import os
import sys

sys.path.insert(1, os.path.join(sys.path[0], ".."))

from orso.filters.bloom_filter import BloomFilter, _get_hash_count, _get_size

ITERATIONS: int = 50000


def random_string(width: int = 16):
    import random

    num_chars = ((width + 1) >> 1) << 3  # Convert length to number of bits
    rand_bytes = random.getrandbits(num_chars)  # Generate random bytes
    rand_hex = hex(rand_bytes)[
        2 : width + 2
    ]  # Convert to hex string and truncate to desired length
    return rand_hex


def test_bloom_filter():
    # first we populate the BloomFilter
    tokens = (random_string(48) for i in range(ITERATIONS))
    bf = BloomFilter()
    for token in tokens:
        bf.add(token)

    # then we test. 100% shouldn't match (we use different string lengths)
    # but we're a probabilistic filter so expect some false positives
    # test a lot more cycles than we used to populate
    SCALE_FACTOR = 10
    tokens = (random_string(32) for i in range(ITERATIONS * SCALE_FACTOR))
    collisions = 0
    for token in tokens:
        if token in bf:
            collisions += 1

    # this is approximately 1% false positive rate, we're going to test between
    # 0.5 and 1.5 because this is probabilistic so are unlikely to actually get 1%
    assert (
        (ITERATIONS * SCALE_FACTOR * 0.005) < collisions < (ITERATIONS * SCALE_FACTOR * 0.015)
    ), collisions / (ITERATIONS * SCALE_FACTOR)


def test_bloom_contains():
    bf = BloomFilter()
    bf.add("test")
    assert "test" in bf
    assert "nonexistent" not in bf


def test_get_size():
    assert _get_size(50000, 0.01) == 479253, _get_size(50000, 0.01)
    assert _get_size(100000, 0.1) == 479253


def test_get_hash_count():
    assert _get_hash_count(479253, 50000) == 7, _get_hash_count(479253, 50000)
    assert _get_hash_count(479253, 100000) == 3


def test_init():
    bf = BloomFilter()
    assert bf.filter_size == 479253
    assert bf.hash_count == 7
    assert len(bf.hash_seeds) == 7
    assert bf.bits.size == 479253, bf.bits.size
    assert sum(bf.bits.array) == 0


if __name__ == "__main__":  # prgama: nocover
    from tests import run_tests

    run_tests()



import random

from orso.cityhash import CityHash64


class CuckooFilter:
    def __init__(self, capacity: int = 10000, bucket_size: int = 2, fingerprint_size: int = 2):
        self.capacity = capacity
        self.bucket_size = bucket_size
        self.fingerprint_size = fingerprint_size
        self.buckets: list = [[] for _ in range(capacity)]

    def __len__(self):
        return sum(len(bucket) for bucket in self.buckets)

    def add(self, item):
        item = str(item)

        hash32 = CityHash64(item)
        hash16_1 = hash32 & 0xFFFFFFFF
        hash16_2 = hash32 >> 32

        # Try to insert fingerprint into first bucket
        bucket_index = hash16_1 % self.capacity
        hash16_1 &= 0xFFFF
        if hash16_1 not in self.buckets[bucket_index]:
            self.buckets[bucket_index].append(hash16_1)
            return True

        # Try to insert fingerprint into second bucket
        bucket_index = hash16_2 % self.capacity
        hash16_2 &= 0xFFFF
        if hash16_2 not in self.buckets[bucket_index]:
            self.buckets[bucket_index].append(hash16_2)
            return True

        # If both buckets are full, perform eviction
        fingerprint = self.buckets[bucket_index][
            random.randint(0, len(self.buckets[bucket_index]) - 1)
        ]
        self.buckets[bucket_index].remove(fingerprint)

        # Try to insert evicted fingerprint
        return self.add(fingerprint)

    def __contains__(self, item):
        item = str(item)

        hash32 = CityHash64(item)
        hash16_1 = hash32 & 0xFFFFFFFF
        hash16_2 = hash32 >> 32

        bucket_index = hash16_1 % self.capacity
        hash16_1 &= 0xFFFF
        if hash16_1 in self.buckets[bucket_index]:
            return True

        bucket_index = hash16_2 % self.capacity
        hash16_2 &= 0xFFFF
        if hash16_2 in self.buckets[bucket_index]:
            return True

        return False


if __name__ == "__main__":  # pragma: no cover
    import os
    import sys

    sys.path.insert(1, os.path.join(sys.path[0], "../.."))

    ITERATIONS = 100000
    import time

    from opteryx.utils import random_string

    from orso.filters import BloomFilter

    t = time.monotonic_ns()
    tokens = (random_string(48) for i in range(ITERATIONS))
    cf = CuckooFilter()
    for token in tokens:
        cf.add(token)

    print((time.monotonic_ns() - t) / 1e9, "cf")

    # then we test. 100% shouldn't match (we use different string lengths)
    # but we're a probabilistic filter so expect some false positives
    # we're configured for a 1% false positive rate
    tokens = (random_string(32) for i in range(ITERATIONS))
    collisions = 0
    for token in tokens:
        if token in cf:
            collisions += 1

    print(ITERATIONS, collisions, (time.monotonic_ns() - t) / 1e9)

    t = time.monotonic_ns()
    tokens = (random_string(48) for i in range(ITERATIONS))
    bf = BloomFilter(10000)
    for token in tokens:
        bf.add(token)

    print((time.monotonic_ns() - t) / 1e9)

    # then we test. 100% shouldn't match (we use different string lengths)
    # but we're a probabilistic filter so expect some false positives
    # we're configured for a 1% false positive rate
    tokens = (random_string(32) for i in range(ITERATIONS))
    collisions = 0
    for token in tokens:
        if token in bf:
            collisions += 1

    print(ITERATIONS, collisions, (time.monotonic_ns() - t) / 1e9)



import os
import sys

import pytest

sys.path.insert(1, os.path.join(sys.path[0], ".."))

from orso.exceptions import DataValidationError
from orso.schema import RelationSchema
from tests import cities


def test_find_column():
    column = cities.schema.find_column("language")
    assert column.name == "language", column.name

    column = cities.schema.find_column("geolocation")
    assert column is None, column


def test_all_column_names():
    column_names = cities.schema.all_column_names()
    assert "name" in column_names
    assert "language" in column_names
    assert "geolocation" not in column_names


def test_schema_persistance():
    as_dict = cities.schema.to_dict()
    from_dict = RelationSchema.from_dict(as_dict)

    assert from_dict == cities.schema
    assert as_dict == from_dict.to_dict()


def test_validate_with_valid_data():
    # Test with valid data
    data = {
        "name": "New York",
        "population": 8623000,
        "country": "United States",
        "founded": "1624",
        "area": 783.8,
        "language": "English",
    }
    assert cities.schema.validate(data) == True


def test_validate_with_missing_column():
    # Test with missing column
    data = {
        "name": "London",
        "population": 8908081,
        "country": "United Kingdom",
        "founded": "43 AD",
        "language": "English",
    }
    with pytest.raises(DataValidationError):
        cities.schema.validate(data)


def test_validate_with_nullable_column():
    # Test with nullable column and value is None
    data = {
        "name": "Paris",
        "population": 2187526,
        "country": "France",
        "founded": None,
        "area": 105.4,
        "language": "French",
    }
    assert cities.schema.validate(data)


def test_validate_with_non_nullable_column():
    # Test with non-nullable column and value is None
    data = {
        "name": None,
        "population": 13929286,
        "country": "Japan",
        "founded": "1457",
        "area": 2187.66,
        "language": "Japanese",
    }
    with pytest.raises(DataValidationError):
        cities.schema.validate(data)


def test_validate_with_wrong_type():
    # Test with column value of wrong type
    data = {
        "name": "Berlin",
        "population": 3769495,
        "country": "Germany",
        "founded": "1237",
        "area": "891.8",  # Expected type is double
        "language": "German",
    }
    with pytest.raises(DataValidationError):
        cities.schema.validate(data)


def test_validate_with_invalid_data_type():
    # Test with invalid data type (not a MutableMapping)
    data = [1, 2, 3]
    with pytest.raises(TypeError):
        cities.schema.validate(data)


def test_schema_iterations():
    schema = cities.schema

    assert schema.num_columns == 6

    for i, column in enumerate(schema.columns):
        assert column == schema.column(i)
        assert column == schema.column(column.name)


if __name__ == "__main__":  # prgama: nocover
    from tests import run_tests

    run_tests()



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import decimal
from functools import wraps
from random import getrandbits


def retry(
    max_tries: int = 3,
    backoff_seconds: int = 1,
    exponential_backoff: bool = False,
    max_backoff: int = 4,
):
    """
    max_tries: int=3, the number of times to execute the function
    backoff_seconds: int=1, the amount of time between retries
    exponential_backoff: bool=False, double the backoff period between retries
    max_backoff: int=4, the maximum backoff (not including the first)
    """

    def decorator_retry(func):
        @wraps(func)
        def wrapper_retry(*args, **kwargs):
            import time

            tries = 0
            this_delay = backoff_seconds
            while tries < max_tries:
                try:
                    return func(*args, **kwargs)
                except Exception as e:
                    tries += 1
                    if tries == max_tries:
                        print(
                            f"`{func.__name__}` failed with `{type(e).__name__}` error, attempt {tries} of {max_tries}. Aborting."
                        )
                        raise e
                    print(
                        f"`{func.__name__}` failed with `{type(e).__name__}` error, attempt {tries} of {max_tries}. Will retry in {this_delay} seconds."
                    )
                    time.sleep(this_delay)
                    if exponential_backoff:
                        this_delay *= 2
                        this_delay = min(this_delay, max_backoff)

        return wrapper_retry

    return decorator_retry


def timed(func):
    """
    Run and report the time for a function
    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        import time

        start_time = time.monotonic_ns()
        result = func(*args, **kwargs)
        end_time = time.monotonic_ns()
        print(f"Function {func.__name__} took {(end_time - start_time)/1e9} seconds to run.")
        return result

    return wrapper


def log(func):
    """
    Log when a function starts and finishes.
    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        import datetime
        import uuid

        unique_marker = str(uuid.uuid4())

        print(f"{datetime.datetime.now()} - Executing {func.__name__} - {unique_marker}")
        result = func(*args, **kwargs)
        print(f"{datetime.datetime.now()} - Finished executing {func.__name__} - {unique_marker}")
        return result

    return wrapper


def repeat(number_of_times: int = None, capture_results: bool = False):
    """
    Repeat a function a given number of times
    """
    if not isinstance(number_of_times, int) or number_of_times < 0:
        raise ValueError("@repeat requires a 'number_of_times' to be set to a positive integer")

    def decorate(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            results = []
            for _ in range(number_of_times):
                result = func(*args, **kwargs)
                if capture_results:
                    results.append(result)
            if capture_results:
                return results
            return None

        return wrapper

    return decorate


def counter(func):
    """
    Add an execution counter and timer to a function
    """
    import time

    def report(self):
        stats = f"\nExecution Statistics for `{self.__name__}`\n  Count   : {self.count}\n"
        if self.count > 0:
            stats += f"  Average : {sum(self._run_times) / self.count} seconds\n"
            stats += f"  Slowest : {min(self._run_times)} seconds\n"
            stats += f"  Fastest : {max(self._run_times)} seconds\n"
        return stats

    @wraps(func)
    def wrapper(*args, **kwargs):
        wrapper.count += 1  # type:ignore
        start_time = time.monotonic()
        result = func(*args, **kwargs)
        wrapper._run_times.append(time.monotonic() - start_time)  # type:ignore
        return result

    wrapper.count = 0  # type:ignore
    wrapper._run_times = []  # type:ignore
    wrapper.stats = lambda: report(wrapper)  # type:ignore
    return wrapper


def throttle(calls_per_second: float):
    import time

    if calls_per_second <= 0:
        raise ValueError("@throttle requires calls_per_second to be greater than zero")
    min_interval = 1.0 / float(calls_per_second)

    def decorate(func):
        last_time_called = 0.0

        @wraps(func)
        def rate_limited_function(*args, **kargs):
            elapsed = time.monotonic() - last_time_called
            left_to_wait = min_interval - elapsed
            if left_to_wait > 0:
                time.sleep(left_to_wait)
            ret = func(*args, **kargs)
            last_time_called = time.monotonic()
            return ret

        return rate_limited_function

    return decorate


def monitor(time_limit=10, interval=1):
    import os
    import threading
    import time

    from orso.exceptions import MissingDependencyError

    try:
        import psutil
    except ImportError as import_error:
        raise MissingDependencyError(import_error.name) from import_error

    def decorator(func):
        stop_flag = False

        def wrapper(*args, **kwargs):
            nonlocal stop_flag
            stop_flag = False

            # Use a thread to monitor the resource usage
            def monitor():
                peak_cpu = 0
                peak_memory = 0
                cpu_tracker = []
                memory_tracker = []

                process = psutil.Process(os.getpid())
                while not stop_flag:
                    cpu_percent = process.cpu_percent(interval=interval)
                    memory_info = process.memory_info().rss

                    cpu_tracker.append(cpu_percent)
                    memory_tracker.append(memory_info)
                    if len(cpu_tracker) > time_limit:
                        cpu_tracker.pop(0)
                        memory_tracker.pop(0)

                    if cpu_percent > peak_cpu:
                        peak_cpu = cpu_percent
                    if memory_info > peak_memory:
                        peak_memory = memory_info

                print(f"Peak CPU usage: {peak_cpu:.2f}%")
                print(f"Peak memory usage: {peak_memory/1024/1024:.2f} MB")

            monitor_thread = threading.Thread(target=monitor)
            monitor_thread.start()

            try:
                start_time = time.monotonic_ns()
                result = func(*args, **kwargs)
                end_time = time.monotonic_ns()
                print(f"Execution time: {(end_time - start_time)/1e9:.6f} seconds")
                return result
            except Exception as e:
                print(f"Error raised: {type(e).__name__}")
                end_time = time.monotonic_ns()
                print(f"Execution time: {(end_time - start_time)/1e9:.6f} seconds")
                raise e
            finally:
                stop_flag = True
                monitor_thread.join()

        return wrapper

    return decorator


def islice(iterator, size):
    for i in range(size):
        yield next(iterator)


class DecimalFactory(decimal.Decimal):
    scale = 0
    precision = 0

    def __call__(self, value):
        context = decimal.Context(prec=self.precision)
        quantized_value = context.create_decimal(value).quantize(decimal.Decimal(10) ** -self.scale)
        return decimal.Decimal(quantized_value)

    def __str__(self):
        return f"Decimal({self.scale},{self.precision})"

    @classmethod
    def new_factory(cls, precision: int, scale: int):
        factory = DecimalFactory.__new__(cls)
        factory.scale = scale
        factory.precision = precision
        return factory


def parquet_type_map(parquet_type):
    import datetime

    from orso.exceptions import MissingDependencyError

    try:
        import pyarrow.lib as lib
    except ImportError as import_error:
        raise MissingDependencyError(import_error.name) from import_error

    if parquet_type.id == lib.Type_NA:
        return None
    if parquet_type.id == lib.Type_BOOL:
        return bool
    if parquet_type.id in {lib.Type_STRING, lib.Type_LARGE_STRING}:
        return str
    if parquet_type.id in {
        lib.Type_INT8,
        lib.Type_INT16,
        lib.Type_INT32,
        lib.Type_INT64,
        lib.Type_UINT8,
        lib.Type_UINT16,
        lib.Type_UINT32,
        lib.Type_UINT64,
    }:
        return int
    if parquet_type.id in {lib.Type_HALF_FLOAT, lib.Type_FLOAT, lib.Type_DOUBLE}:
        return float
    if parquet_type.id in {lib.Type_DECIMAL128, lib.Type_DECIMAL256}:
        return DecimalFactory.new_factory(parquet_type.precision, parquet_type.scale)
    if parquet_type.id in {lib.Type_DATE32}:
        return datetime.date
    if parquet_type.id in {lib.Type_DATE64, 18}:  # not sure what 18 maps to
        return datetime.datetime
    if parquet_type.id in {lib.Type_TIME32, lib.Type_TIME64}:
        return datetime.time
    if parquet_type.id in {lib.Type_INTERVAL_MONTH_DAY_NANO, lib.Type_DURATION}:
        return datetime.timedelta
    if parquet_type.id in {lib.Type_LIST, lib.Type_LARGE_LIST, lib.Type_FIXED_SIZE_LIST}:
        return list
    if parquet_type.id in {lib.Type_STRUCT, lib.Type_MAP}:
        return dict
    if parquet_type.id in {lib.Type_BINARY, lib.Type_LARGE_BINARY}:
        return bytes
    # _UNION_TYPES = {lib.Type_SPARSE_UNION, lib.Type_DENSE_UNION}
    raise ValueError(f"Unable to map parquet type {parquet_type} ({parquet_type.id})")


def random_int() -> int:
    """
    Select a random integer (32bit)
    """
    return getrandbits(32)


def random_string(width: int = 16):
    """
    I've not been able to find a faster way to generate short random strings.
    """
    num_chars = ((width + 1) >> 1) << 3  # Convert length to number of bits
    rand_bytes = getrandbits(num_chars)  # Generate random bytes
    # Convert to hex string and truncate to desired length, we clip the '0x'
    # the start of the hex string, and zero-fill the start of the string
    return ("000000" + hex(rand_bytes)[2:])[-width:]



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class MissingDependencyError(Exception):
    def __init__(self, dependency):
        self.dependency = dependency
        message = f"No module named '{dependency}' can be found, "
        "please install or include in requirements.txt"
        super().__init__(message)


class DataError(Exception):
    pass


class DataValidationError(DataError):
    def __init__(self, column, value, error):
        self.field = column.name
        self.expected_type = column.type
        self.value = value
        self.nullable = column.nullable
        self.error = error
        message = (
            f"Data did not pass validation checks; field `{self.field}` "
            f"with value `{str(value)[16:]}` did not pass "
            f"`{column.type}` {'(nullable)' if self.nullable else ''} check. "
            f"({error})"
        )
        super().__init__(message)


class ColumnDefinitionError(DataError):
    def __init__(self, attribute):
        self.attribute = attribute
        message = f"Column is missing attribute {attribute}"
